from typing import (  # noqa: E501
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from motor import core
from pymongo.collection import Collection

from mongorepository.models import MongoBaseModel

T = TypeVar("T", bound=MongoBaseModel)


class AbstractRepository(Generic[T]):
    def __init__(self, database):
        self.__database = database
        self.__collection_name = self.Config.collection
        self._paginated = self.__get_pagination_flag()
        self._query_limit = self.__get_query_limit_documents()
        self._model_class = self.__orig_bases__[0].__args__[0]

    def __get_pagination_flag(self) -> bool:
        if hasattr(self.Config, "pagination"):
            return self.Config.pagination or False
        return False

    def __get_query_limit_documents(self) -> int:
        if hasattr(self.Config, "pagination"):
            if hasattr(self.Config, "limit"):
                return self.Config.limit
            return 50

    def set_pagination(self, value: bool) -> None:
        self._paginated = value
        self._query_limit = 50

    def get_collection(self) -> Union[Collection, core.Collection]:
        return self.__database.get_collection(self.__collection_name)

    def get_projection(self) -> Dict[str, Any]:
        projection = self._model_class.projection()
        if hasattr(self.Config, "projection"):
            projection = self.Config.projection
        return projection

    def _convert_paginated_results_to_model(
        self, document: Dict[str, Any]
    ) -> None:  # noqa: E501
        document["results"] = [
            self._model_class(**document) for document in document["results"]
        ]

    def generate_pagination_query(
        self,
        query: Dict[str, Any],
        sort: Optional[List[Tuple[str, int]]] = None,
        next_key: Optional[Dict[str, Any]] = None,
    ) -> Tuple[Dict[str, Any], Callable]:
        sort_field = sort[0][0] if sort else None

        def next_key_fn(items):
            if not items:
                return None
            item = items[-1]
            if sort_field is None:
                return {"_id": item["_id"]}

            return {"_id": item["_id"], sort_field: item[sort_field]}

        if next_key is None:
            return query, next_key_fn

        paginated_query = query.copy()

        if sort is None:
            paginated_query["_id"] = {"$gt": next_key["_id"]}
            return paginated_query, next_key_fn

        sort_operator = "$gt" if sort[0][1] == 1 else "$lt"

        pagination_query = [
            {sort_field: {sort_operator: next_key[sort_field]}},
            {
                "$and": [
                    {sort_field: next_key[sort_field]},
                    {"_id": {sort_operator: next_key["_id"]}},
                ]
            },
        ]

        if "$or" not in paginated_query:
            paginated_query["$or"] = pagination_query
        else:
            paginated_query = {"$and": [query, {"$or": pagination_query}]}

        return paginated_query, next_key_fn



from datetime import datetime, timezone
from typing import Any, Dict, Optional, TypeVar

from pydantic import BaseModel, Field

from mongorepository.utils.objects import PyObjectId

T = TypeVar("T", bound=BaseModel)


def date_tzinfo():
    return datetime.now().replace(tzinfo=timezone.utc)


class MongoBaseModel(BaseModel):
    id: Optional[PyObjectId] = Field(alias="_id")
    created: datetime = Field(default_factory=date_tzinfo)
    updated: datetime = Field(default_factory=date_tzinfo)

    def update_from_model(self, model: T) -> None:
        updates = model.dict(exclude_none=True)
        fields = updates.keys()
        for field in fields:
            setattr(self, field, updates[field])

    @classmethod
    def projection(cls) -> Dict[str, Any]:
        fields = cls.__fields__
        keys = fields.keys()
        mapper = {}

        for key in keys:
            value = fields[key]
            if value.alias:
                mapper[value.alias] = 1
            else:
                mapper[key] = 1

        return mapper



from typing import Any, Dict, List, Optional, Tuple, Union

import pymongo
from bson import ObjectId
from pymongo.database import Database
from pymongo.results import DeleteResult, InsertManyResult, InsertOneResult

from mongorepository.repositories.base import AbstractRepository, T


class Repository(AbstractRepository[T]):
    def __init__(self, database: Database):
        super().__init__(database)

    def __get_paginated_documents(
        self,
        query: Dict[str, Any],
        sort: List[Tuple[str, int]],
        next_key=Dict[str, Any],
        projection=Dict[str, Any],
    ) -> Dict[str, Any]:  # noqa: E501
        query, next_key_fn = self.generate_pagination_query(
            query, sort, next_key
        )  # noqa: E501

        cursor = (
            self.get_collection()
            .find(query, projection=projection or self.get_projection())
            .sort(sort)
            .limit(self._query_limit)
        )

        documents = [document for document in cursor]  # noqa: E501

        return {
            "total": len(documents),
            "results": documents,
            "next_page": next_key_fn(documents),
        }

    def list_objects(
        self,
        query: Optional[dict] = None,
        sort: Optional[list] = None,
        next_page: Optional[dict] = None,
        projection: Optional[dict] = None,
    ) -> Union[List[T], Dict[str, Any]]:
        collection = self.get_collection()
        if query is None:
            query = {}

        if not sort:
            sort = [("_id", pymongo.DESCENDING)]

        if self._paginated:
            result = self.__get_paginated_documents(
                query, sort, next_page, projection or self.get_projection()
            )
            self._convert_paginated_results_to_model(result)
            return result

        cursor = collection.find(
            query, projection=projection or self.get_projection()
        ).sort(sort)

        return [self._model_class(**document) for document in cursor]

    def find_by_query(
        self, query: dict, projection: Optional[dict] = None
    ) -> Optional[T]:
        collection = self.get_collection()
        if document := collection.find_one(
            query, projection=projection or self.get_projection()
        ):
            return self._model_class(**document)
        return None

    def find_by_id(
        self, document_id: str, projection: Optional[dict] = None
    ) -> Optional[T]:
        collection = self.get_collection()
        if document := collection.find_one(
            {"_id": ObjectId(document_id)},
            projection=projection or self.get_projection(),
        ):
            return self._model_class(**document)
        return None

    def save(self, model: T) -> Optional[T]:
        collection = self.get_collection()
        raw_model = model.dict(by_alias=True, exclude_none=True)

        if id_model := raw_model.get("_id", raw_model.get("id")):
            collection.update_one(
                {"_id": ObjectId(id_model)}, {"$set": raw_model}
            )  # noqa: E501
            return self.find_by_id(model.id)

        document: InsertOneResult = collection.insert_one(raw_model)

        return self.find_by_id(str(document.inserted_id))

    def bulk_create(self, models: List[T]) -> List[ObjectId]:
        raw_models = [model.dict(exclude_none=True) for model in models]
        result: InsertManyResult = self.get_collection().insert_many(
            raw_models
        )  # noqa: E501
        return result.inserted_ids

    def delete(self, object_id: Union[str, ObjectId]) -> bool:
        collection = self.get_collection()
        if isinstance(object_id, str):
            object_id = ObjectId(object_id)

        result: DeleteResult = collection.delete_one({"_id": object_id})

        if result.deleted_count == 1:
            return True
        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mongorepository', 'mongorepository.repositories', 'mongorepository.utils']

package_data = \
{'': ['*']}

install_requires = \
['motor>=3.1.1,<4.0.0', 'pydantic>=1.10.6,<2.0.0']

setup_kwargs = {
    'name': 'mongorepository',
    'version': '0.4.2',
    'description': '',
    'long_description': '# Mongo Repository\n\nThis package provide a sync and async repositories utilities for mongodb.',
    'author': 'Ramon Rodrigues',
    'author_email': 'ramon.srodrigues01@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
